@@ -32,6 +32,7 @@
 #include <connman/plugin.h>
 #include <connman/device.h>
 #include <connman/network.h>
+#include <connman/service.h>
 #include <connman/inet.h>
 #include <connman/dbus.h>
 
@@ -86,17 +87,20 @@ static char *get_ident(const char *path)
 	return pos + 1;
 }
 
-static void create_device(struct dundee_data *info)
+static int create_device(struct dundee_data *info)
 {
 	struct connman_device *device;
 	char *ident;
+	int err;
 
 	DBG("%s", info->path);
 
 	ident = g_strdup(get_ident(info->path));
 	device = connman_device_create(ident, CONNMAN_DEVICE_TYPE_BLUETOOTH);
-	if (device == NULL)
+	if (device == NULL) {
+		err = -ENOMEM;
 		goto out;
+	}
 
 	DBG("device %p", device);
 
@@ -106,7 +110,8 @@ static void create_device(struct dundee_data *info)
 
 	connman_device_set_data(device, info);
 
-	if (connman_device_register(device) < 0) {
+	err = connman_device_register(device);
+	if (err < 0) {
 		connman_error("Failed to register DUN device");
 		connman_device_unref(device);
 		goto out;
@@ -116,6 +121,7 @@ static void create_device(struct dundee_data *info)
 
 out:
 	g_free(ident);
+	return err;
 }
 
 static void destroy_device(struct dundee_data *info)
@@ -150,17 +156,18 @@ static void device_destroy(gpointer data)
 	g_free(info);
 }
 
-static void create_network(struct dundee_data *info)
+static int create_network(struct dundee_data *info)
 {
 	struct connman_network *network;
 	const char *group;
+	int err;
 
 	DBG("%s", info->path);
 
 	network = connman_network_create(info->path,
 				CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN);
 	if (network == NULL)
-		return;
+		return -ENOMEM;
 
 	DBG("network %p", network);
 
@@ -176,20 +183,30 @@ static void create_network(struct dundee_data *info)
 
 	connman_network_set_available(network, TRUE);
 
-	if (connman_device_add_network(info->device, network) < 0) {
+	err = connman_device_add_network(info->device, network);
+	if (err < 0) {
 		connman_network_unref(network);
-		return;
+		return err;
 	}
 
 	info->network = network;
+
+	return 0;
 }
 
 static void set_connected(struct dundee_data *info)
 {
+	struct connman_service *service;
+
 	DBG("%s", info->path);
 
 	connman_inet_ifup(info->index);
 
+	service = connman_service_lookup_from_network(info->network);
+	if (service == NULL)
+		return;
+
+	connman_service_create_ip4config(service, info->index);
 	connman_network_set_index(info->network, info->index);
 	connman_network_set_ipv4_method(info->network,
 					CONNMAN_IPCONFIG_METHOD_FIXED);
@@ -556,6 +573,7 @@ static gboolean device_changed(DBusConnection *conn,
 static void add_device(const char *path, DBusMessageIter *properties)
 {
 	struct dundee_data *info;
+	int err;
 
 	info = g_hash_table_lookup(dundee_devices, path);
 	if (info != NULL)
@@ -601,11 +619,23 @@ static void add_device(const char *path, DBusMessageIter *properties)
 
 	g_hash_table_insert(dundee_devices, g_strdup(path), info);
 
-	create_device(info);
-	create_network(info);
+	err = create_device(info);
+	if (err < 0)
+		goto out;
+
+	err = create_network(info);
+	if (err < 0) {
+		destroy_device(info);
+		goto out;
+	}
 
 	if (info->active == TRUE)
 		set_connected(info);
+
+	return;
+
+out:
+	g_hash_table_remove(dundee_devices, path);
 }
 
 static gboolean device_added(DBusConnection *conn, DBusMessage *message,
