@@ -31,6 +31,7 @@
 #include <sys/inotify.h>
 #include <glib.h>
 
+#include <connman/provision.h>
 #include "connman.h"
 
 struct connman_config_service {
@@ -51,6 +52,7 @@ struct connman_config_service {
 	GSList *service_identifiers;
 	char *config_ident; /* file prefix */
 	char *config_entry; /* entry name */
+	connman_bool_t hidden;
 };
 
 struct connman_config {
@@ -89,6 +91,7 @@ static connman_bool_t cleanup = FALSE;
 #define SERVICE_KEY_IDENTITY           "Identity"
 #define SERVICE_KEY_PHASE2             "Phase2"
 #define SERVICE_KEY_PASSPHRASE         "Passphrase"
+#define SERVICE_KEY_HIDDEN             "Hidden"
 
 static const char *config_possible_keys[] = {
 	CONFIG_KEY_NAME,
@@ -110,6 +113,7 @@ static const char *service_possible_keys[] = {
 	SERVICE_KEY_IDENTITY,
 	SERVICE_KEY_PHASE2,
 	SERVICE_KEY_PASSPHRASE,
+	SERVICE_KEY_HIDDEN,
 	NULL,
 };
 
@@ -387,13 +391,16 @@ static int load_service(GKeyFile *keyfile, const char *group,
 	service->config_ident = g_strdup(config->ident);
 	service->config_entry = g_strdup_printf("service_%s", service->ident);
 
+	service->hidden = g_key_file_get_boolean(keyfile, group,
+						SERVICE_KEY_HIDDEN, NULL);
+
 	if (service_created)
 		g_hash_table_insert(config->service_table, service->ident,
 					service);
 
 	if (config->protected == TRUE)
 		protected_services =
-			g_slist_append(protected_services, service);
+			g_slist_prepend(protected_services, service);
 
 	connman_info("Adding service configuration %s", service->ident);
 
@@ -448,6 +455,7 @@ static int load_config(struct connman_config *config)
 		config->protected = protected;
 	else
 		config->protected = TRUE;
+	g_clear_error(&error);
 
 	groups = g_key_file_get_groups(keyfile, &length);
 
@@ -849,6 +857,9 @@ static void provision_service(gpointer key, gpointer value, gpointer user_data)
 	if (config->passphrase != NULL)
 		__connman_service_set_string(service, "Passphrase", config->passphrase);
 
+	if (config->hidden == TRUE)
+		__connman_service_set_hidden(service);
+
 	__connman_service_mark_dirty();
 
 	__connman_service_save(service);
@@ -894,7 +905,7 @@ int __connman_config_provision_service_ident(struct connman_service *service,
 		return -ENOSYS;
 
 	config = g_hash_table_lookup(config_table, ident);
-	if(config != NULL) {
+	if (config != NULL) {
 		GHashTableIter iter;
 		gpointer value, key;
 		gboolean found = FALSE;
@@ -908,15 +919,20 @@ int __connman_config_provision_service_ident(struct connman_service *service,
 		if (file != NULL && entry != NULL) {
 			while (g_hash_table_iter_next(&iter, &key,
 							&value) == TRUE) {
-				struct connman_config_service *config = value;
+				struct connman_config_service *config_service;
 
-				if (g_strcmp0(config->config_ident,
-								file) == 0 &&
-						g_strcmp0(config->config_entry,
-								entry) == 0) {
-					found = TRUE;
-					break;
-				}
+				config_service = value;
+
+				if (g_strcmp0(config_service->config_ident,
+								file) != 0)
+					continue;
+
+				if (g_strcmp0(config_service->config_entry,
+								entry) != 0)
+					continue;
+
+				found = TRUE;
+				break;
 			}
 
 			DBG("found %d ident %s file %s entry %s", found, ident,
@@ -938,3 +954,83 @@ int __connman_config_provision_service_ident(struct connman_service *service,
 
 	return ret;
 }
+
+struct connman_config_entry **connman_config_get_entries(void)
+{
+	GHashTableIter iter_file, iter_config;
+	gpointer value, key;
+	struct connman_config_entry **entries = NULL;
+	int i = 0, count;
+
+	g_hash_table_iter_init(&iter_file, config_table);
+	while (g_hash_table_iter_next(&iter_file, &key, &value) == TRUE) {
+		struct connman_config *config_file = value;
+
+		count = g_hash_table_size(config_file->service_table);
+
+		entries = g_try_realloc(entries, (i + count + 1) *
+					sizeof(struct connman_config_entry *));
+		if (entries == NULL)
+			return NULL;
+
+		g_hash_table_iter_init(&iter_config,
+						config_file->service_table);
+		while (g_hash_table_iter_next(&iter_config, &key,
+							&value) == TRUE) {
+			struct connman_config_service *config = value;
+
+			entries[i] = g_try_new0(struct connman_config_entry,
+						1);
+			if (entries[i] == NULL)
+				goto cleanup;
+
+			entries[i]->ident = g_strdup(config->ident);
+			entries[i]->name = g_strdup(config->name);
+			entries[i]->ssid = g_try_malloc0(config->ssid_len + 1);
+			if (entries[i]->ssid == NULL)
+				goto cleanup;
+
+			memcpy(entries[i]->ssid, config->ssid,
+							config->ssid_len);
+			entries[i]->ssid_len = config->ssid_len;
+			entries[i]->hidden = config->hidden;
+
+			i++;
+		}
+	}
+
+	if (entries != NULL) {
+		entries = g_try_realloc(entries, (i + 1) *
+					sizeof(struct connman_config_entry *));
+		if (entries == NULL)
+			return NULL;
+
+		entries[i] = NULL;
+
+		DBG("%d provisioned AP found", i);
+	}
+
+	return entries;
+
+cleanup:
+	connman_config_free_entries(entries);
+	return NULL;
+}
+
+void connman_config_free_entries(struct connman_config_entry **entries)
+{
+	int i;
+
+	if (entries == NULL)
+		return;
+
+	for (i = 0; entries[i]; i++) {
+		g_free(entries[i]->ident);
+		g_free(entries[i]->name);
+		g_free(entries[i]->ssid);
+		g_free(entries[i]);
+	}
+
+	g_free(entries);
+	return;
+}
