@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -169,9 +169,6 @@ int connman_ipaddress_set_ipv6(struct connman_ipaddress *ipaddress,
 	if (check_ipv6_address(address) == FALSE)
 		return -EINVAL;
 
-	if (check_ipv6_address(gateway) == FALSE)
-		return -EINVAL;
-
 	DBG("prefix_len %d address %s gateway %s",
 			prefix_length, address, gateway);
 
@@ -635,8 +632,19 @@ void __connman_ipconfig_newlink(int index, unsigned short type,
 		return;
 
 	ipdevice = g_hash_table_lookup(ipdevice_hash, GINT_TO_POINTER(index));
-	if (ipdevice != NULL)
+	if (ipdevice != NULL) {
+		char *ifname = connman_inet_ifname(index);
+		if (g_strcmp0(ipdevice->ifname, ifname) != 0) {
+			DBG("interface name changed %s -> %s",
+				ipdevice->ifname, ifname);
+
+			g_free(ipdevice->ifname);
+			ipdevice->ifname = ifname;
+		} else
+			g_free(ifname);
+
 		goto update;
+	}
 
 	ipdevice = g_try_new0(struct connman_ipdevice, 1);
 	if (ipdevice == NULL)
@@ -810,7 +818,7 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 	else
 		return;
 
-	ipdevice->address_list = g_slist_append(ipdevice->address_list,
+	ipdevice->address_list = g_slist_prepend(ipdevice->address_list,
 								ipaddress);
 
 	connman_info("%s {add} address %s/%u label %s family %d",
@@ -1245,6 +1253,7 @@ void __connman_ipconfig_set_prefixlen(struct connman_ipconfig *ipconfig, unsigne
 static struct connman_ipconfig *create_ipv6config(int index)
 {
 	struct connman_ipconfig *ipv6config;
+	struct connman_ipdevice *ipdevice;
 
 	DBG("index %d", index);
 
@@ -1258,7 +1267,10 @@ static struct connman_ipconfig *create_ipv6config(int index)
 	ipv6config->enabled = FALSE;
 	ipv6config->type = CONNMAN_IPCONFIG_TYPE_IPV6;
 	ipv6config->method = CONNMAN_IPCONFIG_METHOD_AUTO;
-	ipv6config->ipv6_privacy_config = 0;
+
+	ipdevice = g_hash_table_lookup(ipdevice_hash, GINT_TO_POINTER(index));
+	if (ipdevice != NULL)
+		ipv6config->ipv6_privacy_config = ipdevice->ipv6_privacy;
 
 	ipv6config->address = connman_ipaddress_alloc(AF_INET6);
 	if (ipv6config->address == NULL) {
@@ -1660,6 +1672,25 @@ void __connman_ipconfig_disable_ipv6(struct connman_ipconfig *ipconfig)
 	disable_ipv6(ipconfig);
 }
 
+connman_bool_t __connman_ipconfig_is_usable(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL)
+		return FALSE;
+
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+		return FALSE;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		break;
+	}
+
+	return TRUE;
+}
+
 int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
@@ -1849,6 +1880,7 @@ static int string2privacy(const char *privacy)
 void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 							DBusMessageIter *iter)
 {
+	struct connman_ipaddress *append_addr = NULL;
 	const char *str;
 
 	DBG("");
@@ -1862,33 +1894,50 @@ void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 
 	connman_dbus_dict_append_basic(iter, "Method", DBUS_TYPE_STRING, &str);
 
-	if (ipconfig->system == NULL)
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		return;
 
-	if (ipconfig->system->local != NULL) {
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+		append_addr = ipconfig->address;
+		break;
+
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+		append_addr = ipconfig->system;
+		break;
+	}
+
+	if (append_addr == NULL)
+		return;
+
+	if (append_addr->local != NULL) {
 		in_addr_t addr;
 		struct in_addr netmask;
 		char *mask;
 
 		connman_dbus_dict_append_basic(iter, "Address",
-				DBUS_TYPE_STRING, &ipconfig->system->local);
+				DBUS_TYPE_STRING, &append_addr->local);
 
-		addr = 0xffffffff << (32 - ipconfig->system->prefixlen);
+		addr = 0xffffffff << (32 - append_addr->prefixlen);
 		netmask.s_addr = htonl(addr);
 		mask = inet_ntoa(netmask);
 		connman_dbus_dict_append_basic(iter, "Netmask",
 						DBUS_TYPE_STRING, &mask);
 	}
 
-	if (ipconfig->system->gateway != NULL)
+	if (append_addr->gateway != NULL)
 		connman_dbus_dict_append_basic(iter, "Gateway",
-				DBUS_TYPE_STRING, &ipconfig->system->gateway);
+				DBUS_TYPE_STRING, &append_addr->gateway);
 }
 
 void __connman_ipconfig_append_ipv6(struct connman_ipconfig *ipconfig,
 					DBusMessageIter *iter,
 					struct connman_ipconfig *ipconfig_ipv4)
 {
+	struct connman_ipaddress *append_addr = NULL;
 	const char *str, *privacy;
 
 	DBG("");
@@ -1908,20 +1957,36 @@ void __connman_ipconfig_append_ipv6(struct connman_ipconfig *ipconfig,
 
 	connman_dbus_dict_append_basic(iter, "Method", DBUS_TYPE_STRING, &str);
 
-	if (ipconfig->system == NULL)
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
 		return;
 
-	if (ipconfig->system->local != NULL) {
-		connman_dbus_dict_append_basic(iter, "Address",
-				DBUS_TYPE_STRING, &ipconfig->system->local);
-		connman_dbus_dict_append_basic(iter, "PrefixLength",
-						DBUS_TYPE_BYTE,
-						&ipconfig->system->prefixlen);
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+		append_addr = ipconfig->address;
+		break;
+
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
+		append_addr = ipconfig->system;
+		break;
 	}
 
-	if (ipconfig->system->gateway != NULL)
+	if (append_addr == NULL)
+		return;
+
+	if (append_addr->local != NULL) {
+		connman_dbus_dict_append_basic(iter, "Address",
+				DBUS_TYPE_STRING, &append_addr->local);
+		connman_dbus_dict_append_basic(iter, "PrefixLength",
+						DBUS_TYPE_BYTE,
+						&append_addr->prefixlen);
+	}
+
+	if (append_addr->gateway != NULL)
 		connman_dbus_dict_append_basic(iter, "Gateway",
-				DBUS_TYPE_STRING, &ipconfig->system->gateway);
+				DBUS_TYPE_STRING, &append_addr->gateway);
 
 	privacy = privacy2string(ipconfig->ipv6_privacy_config);
 	connman_dbus_dict_append_basic(iter, "Privacy",
@@ -2302,8 +2367,9 @@ int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 	}
 
 	key = g_strdup_printf("%snetmask_prefixlen", prefix);
-	g_key_file_set_integer(keyfile, identifier,
-			key, ipconfig->address->prefixlen);
+	if (ipconfig->address->prefixlen != 0)
+		g_key_file_set_integer(keyfile, identifier,
+				key, ipconfig->address->prefixlen);
 	g_free(key);
 
 	key = g_strdup_printf("%slocal_address", prefix);
